"""Parser for list types"""
from __future__ import annotations
from typing import Any, List, TYPE_CHECKING, Type, TypeVar, get_args, get_origin

from .abc import ArgSchemaParser

if TYPE_CHECKING:
    from ..json_type import JsonType
    from typing_extensions import TypeGuard

T = TypeVar("T")


class ListParser(ArgSchemaParser[List[T]]):
    """Parser for list types"""

    @classmethod
    def can_parse(cls, argtype: Any) -> TypeGuard[Type[List[T]]]:
        return get_origin(argtype) in [
            list,
            List,
        ]

    @property
    def argument_schema(self) -> dict[str, JsonType]:
        return {
            "type": "array",
            "items": self.parse_rec(get_args(self.argtype)[0]).argument_schema,
        }

    def parse_value(self, value: JsonType) -> List[T]:
        if not isinstance(value, list):
            raise TypeError(f"Expected list, got {value}")
        return [self.parse_rec(get_args(self.argtype)[0]).parse_value(v) for v in value]



"""A module for running OpenAI functions"""
from __future__ import annotations
import json
from typing import TYPE_CHECKING

from .exceptions import FunctionNotFoundError
from .functions import FunctionResult, OpenAIFunction
from .sets import FunctionSet

if TYPE_CHECKING:
    from ..json_type import JsonType
    from ..openai_types import FunctionCall


class BasicFunctionSet(FunctionSet):
    """A skill set"""

    def __init__(
        self,
        functions: list[OpenAIFunction] | None = None,
    ) -> None:
        self.functions = functions or []

    @property
    def functions_schema(self) -> list[JsonType]:
        """Get the functions schema

        Returns:
            JsonType: The functions schema
        """
        return [function.schema for function in self.functions]

    def run_function(self, input_data: FunctionCall) -> FunctionResult:
        """Run the function

        Args:
            input_data (FunctionCall): The function call

        Returns:
            FunctionResult: The function output

        Raises:
            FunctionNotFoundError: If the function is not found
        """
        function = self.find_function(input_data["name"])
        result = self.get_function_result(function, json.loads(input_data["arguments"]))
        return FunctionResult(function.name, result, function.interpret_as_response)

    def find_function(self, function_name: str) -> OpenAIFunction:
        """Find a function

        Args:
            function_name (str): The function name

        Returns:
            OpenAIFunction: The function

        Raises:
            FunctionNotFoundError: If the function is not found
        """
        for function in self.functions:
            if function.name == function_name:
                return function
        raise FunctionNotFoundError(f"Function {function_name} not found")

    def get_function_result(
        self, function: OpenAIFunction, arguments: dict[str, JsonType]
    ) -> str | None:
        """Get the result of a function

        Args:
            function (OpenAIFunction): The function
            arguments (dict[str, JsonType]): The arguments

        Raises:
            TypeError: If the function returns a non-string value
                while serialize is False

        Returns:
            str | None: The result
        """
        result = function(arguments)

        if function.save_return:
            if function.serialize:
                return json.dumps(result)
            if isinstance(result, str):
                return result
            raise TypeError(f"Function {function.name} returned a non-string value")
        return None

    def _add_function(self, function: OpenAIFunction) -> None:
        """Add a function

        Args:
            function (OpenAIFunction): The function
        """
        self.functions.append(function)

    def _remove_function(self, name: str) -> None:
        """Remove a function

        Args:
            name (str): The name of the function to remove
        """
        self.functions = [f for f in self.functions if f.name != name]



"""Parser for enum types"""
from __future__ import annotations
import enum
from typing import Any, TYPE_CHECKING, Type, TypeVar

from .abc import ArgSchemaParser

if TYPE_CHECKING:
    from ..json_type import JsonType
    from typing_extensions import TypeGuard

T = TypeVar("T", bound=enum.Enum)


class EnumParser(ArgSchemaParser[T]):
    """Parser for enum types"""

    @classmethod
    def can_parse(cls, argtype: Any) -> TypeGuard[Type[T]]:
        if not isinstance(argtype, type):
            return False
        return issubclass(argtype, enum.Enum)

    @property
    def argument_schema(self) -> dict[str, JsonType]:
        schema: dict[str, JsonType] = {
            "type": "string",
            "enum": [e.name for e in self.argtype],
        }
        if self.argtype.__doc__ is not None:
            schema["description"] = self.argtype.__doc__
        return schema

    def parse_value(self, value: JsonType) -> T:
        if not isinstance(value, str):
            raise TypeError(f"Expected str, got {value}")
        return self.argtype[value]



"""A set of tools responsible for managing the functions themselves."""
from .basic_set import BasicFunctionSet
from .exceptions import FunctionNotFoundError
from .sets import FunctionSet
from .wrapper import FunctionWrapper, WrapperConfig

__all__ = [
    "BasicFunctionSet",
    "FunctionNotFoundError",
    "FunctionSet",
    "FunctionWrapper",
    "WrapperConfig",
]



"""Parser for dataclass types"""
from __future__ import annotations
import dataclasses
from typing import Any, ClassVar, Protocol, TYPE_CHECKING, Type

from .abc import ArgSchemaParser

if TYPE_CHECKING:
    from ..json_type import JsonType
    from typing_extensions import TypeGuard


class IsDataclass(Protocol):  # pylint: disable=too-few-public-methods
    """A protocol for checking if a class is a dataclass"""

    __dataclass_fields__: ClassVar[dict]


class DataclassParser(ArgSchemaParser[IsDataclass]):
    """Parser for dataclass types"""

    @classmethod
    def can_parse(cls, argtype: Any) -> TypeGuard[Type[IsDataclass]]:
        return dataclasses.is_dataclass(argtype)

    @property
    def argument_schema(self) -> dict[str, JsonType]:
        return {
            "type": "object",
            "description": self.argtype.__doc__,
            "properties": {
                field.name: self.parse_rec(field.type).argument_schema
                for field in dataclasses.fields(self.argtype)
            },
            "required": [
                field.name
                for field in dataclasses.fields(self.argtype)
                if field.default is dataclasses.MISSING
            ],
        }

    def parse_value(self, value: JsonType) -> IsDataclass:
        if not isinstance(value, dict):
            raise TypeError(f"Expected dict, got {value}")
        return self.argtype(
            **{
                field.name: self.parse_rec(field.type).parse_value(value[field.name])
                for field in dataclasses.fields(self.argtype)
                if field.name in value
            }
        )



"""A module for running OpenAI functions"""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING, overload

import openai

from .functions.union import UnionSkillSet
from .openai_types import (
    FinalResponseMessage,
    FunctionMessageType,
    GenericMessage,
    Message,
    is_final_response_message,
)


if TYPE_CHECKING:
    from .json_type import JsonType
    from .openai_types import FunctionCall, MessageType, NonFunctionMessageType
    from .functions.sets import FunctionResult, FunctionSet, OpenAIFunction


class Conversation:
    """A class for running OpenAI functions"""

    def __init__(
        self,
        skills: list[FunctionSet] | None = None,
        model: str = "gpt-3.5-turbo-0613",
    ) -> None:
        self.messages: list[GenericMessage] = []
        self.skills = UnionSkillSet(*(skills or []))
        self.model = model

    @property
    def functions_schema(self) -> JsonType:
        """Get the functions schema

        Returns:
            JsonType: The functions schema
        """
        return self.skills.functions_schema

    def _add_message(self, message: GenericMessage) -> None:
        """Add a message

        Args:
            message (GenericMessage): The message
        """
        self.messages.append(message)

    def add_message(self, message: GenericMessage | MessageType | str) -> None:
        """Add a message

        Args:
            message (GenericMessage | MessageType | str): The message
        """
        if isinstance(message, GenericMessage):
            self._add_message(message)
        else:
            self._add_message(Message(message))

    def add_messages(self, messages: list[GenericMessage | MessageType]) -> None:
        """Add messages

        Args:
            messages (list[GenericMessage | MessageType]): The messages
        """
        for message in messages:
            self.add_message(message)

    def pop_message(self, index: int = -1) -> GenericMessage:
        """Pop a message

        Args:
            index (int): The index. Defaults to -1.

        Returns:
            GenericMessage: The message
        """
        return self.messages.pop(index)

    def clear_messages(self) -> None:
        """Clear the messages"""
        self.messages = []

    def _generate_message(self) -> NonFunctionMessageType:
        """Generate a response

        Returns:
            NonFunctionMessageType: The response
        """
        response = openai.ChatCompletion.create(
            model=self.model,
            messages=[message.as_dict() for message in self.messages],
            functions=self.functions_schema,
            function_call="auto",
        )
        return response["choices"][0]["message"]  # type: ignore

    def substitute_last_with_function_result(self, result: str) -> None:
        """Substitute the last message with the result

        Args:
            result (str): The function result
        """
        self.pop_message()
        response: NonFunctionMessageType = {
            "role": "assistant",
            "content": result,
        }
        self.add_message(response)

    def add_function_result(self, function_result: FunctionResult) -> bool:
        """Add a function result

        Args:
            function_result (FunctionResult): The function result

        Returns:
            bool: Whether the function result was added
        """
        if function_result.content is None:
            return False
        response: FunctionMessageType = {
            "role": "function",
            "name": function_result.name,
            "content": function_result.content,
        }
        self.add_message(response)
        return True

    def add_or_substitute_function_result(
        self, function_result: FunctionResult
    ) -> bool:
        """Add or substitute a function result

        Args:
            function_result (FunctionResult): The function result

        Raises:
            TypeError: If the function returns a None value

        Returns:
            bool: Whether the function result was added
        """
        if function_result.substitute:
            if function_result.content is None:
                raise TypeError(
                    f"Function {function_result.name} returned a None value"
                )
            self.substitute_last_with_function_result(function_result.content)
            return True
        return self.add_function_result(function_result)

    def run_function_and_substitute(
        self,
        function_call: FunctionCall,
    ) -> bool:
        """Run a function, replacing the last message with the result if needed

        Args:
            function_call (FunctionCall): The function call

        Raises:
            TypeError: If the function returns a None value

        Returns:
            bool: Whether the function result was added
        """
        return self.add_or_substitute_function_result(
            self.skills.run_function(function_call)
        )

    def run_function_if_needed(self) -> bool:
        """Run a function if needed

        Returns:
            bool: Whether the function result was added
        """
        if not self.messages:
            return False

        function_call = self.messages[-1].function_call
        if not function_call:
            return False

        return self.run_function_and_substitute(function_call)

    def generate_message(self) -> GenericMessage:
        """Generate the next message

        Returns:
            GenericMessage: The response
        """
        if self.run_function_if_needed():
            return self.messages[-1]

        message = self._generate_message()
        self.add_message(message)
        return Message(message)

    def run_until_response(self) -> FinalResponseMessage:
        """Run until a response is generated

        Returns:
            FinalResponseMessage: The response
        """
        while True:
            message = self.generate_message()
            if is_final_response_message(message):
                return message

    @overload
    def add_function(self, function: OpenAIFunction) -> OpenAIFunction:
        ...

    @overload
    def add_function(
        self,
        function: Callable[..., JsonType],
        *,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> Callable[..., JsonType]:
        ...

    @overload
    def add_function(
        self,
        *,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> Callable[[Callable[..., JsonType]], Callable[..., JsonType]]:
        ...

    def add_function(
        self,
        function: OpenAIFunction | Callable[..., JsonType] | None = None,
        *,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> (
        Callable[[Callable[..., JsonType]], Callable[..., JsonType]]
        | Callable[..., JsonType]
    ):
        """Add a function

        Args:
            function (OpenAIFunction | Callable[..., JsonType]): The function
            save_return (bool): Whether to send the return value of this
                function to the AI. Defaults to True.
            serialize (bool): Whether to serialize the return value of this
                function. Defaults to True. Otherwise, the return value must be a
                string.
            interpret_as_response (bool): Whether to interpret the return
                value of this function as a response of the agent. Defaults to False.

        Returns:
            Callable[[Callable[..., JsonType]], Callable[..., JsonType]]: A decorator
            Callable[..., JsonType]: The original function
        """
        if function is None:
            return self.skills.add_function(
                save_return=save_return,
                serialize=serialize,
                interpret_as_response=interpret_as_response,
            )
        return self.skills.add_function(
            function,
            save_return=save_return,
            serialize=serialize,
            interpret_as_response=interpret_as_response,
        )

    def remove_function(
        self, function: str | OpenAIFunction | Callable[..., JsonType]
    ) -> None:
        """Remove a function

        Args:
            function (str | OpenAIFunction | Callable[..., JsonType]): The function
        """
        self.skills.remove_function(function)

    def ask(self, question: str) -> str:
        """Ask the AI a question

        Args:
            question (str): The question

        Returns:
            str: The answer to the question
        """
        self.add_message(question)
        return self.run_until_response().content

    def add_skill(self, skill: FunctionSet) -> None:
        """Add a skill

        Args:
            skill (FunctionSet): The skill to add
        """
        self.skills.add_skill(skill)



"""ChatGPT function calling based on function docstrings."""
from .conversation import Conversation
from .functions import (
    BasicFunctionSet,
    FunctionNotFoundError,
    FunctionSet,
    FunctionWrapper,
    WrapperConfig,
)
from .openai_types import Message
from .parsers import ArgSchemaParser, defargparsers

__all__ = [
    "Conversation",
    "BasicFunctionSet",
    "FunctionNotFoundError",
    "FunctionSet",
    "defargparsers",
    "ArgSchemaParser",
    "FunctionWrapper",
    "Message",
    "WrapperConfig",
]



"""Wrap a function for jsonschema io."""
from __future__ import annotations
from dataclasses import dataclass
import inspect
from typing import Any, Callable, OrderedDict, TYPE_CHECKING, Type

from docstring_parser import Docstring, parse

from ..parsers import ArgSchemaParser, defargparsers

if TYPE_CHECKING:
    from ..json_type import JsonType


@dataclass
class WrapperConfig:
    """Configuration for a FunctionWrapper"""

    parsers: list[Type[ArgSchemaParser]] | None = None
    save_return: bool = True
    serialize: bool = True
    interpret_as_response: bool = False


class FunctionWrapper:
    """Wraps a function for jsonschema io"""

    def __init__(
        self,
        func: Callable[..., JsonType],
        config: WrapperConfig | None = None,
    ) -> None:
        """Initialize a FunctionWrapper

        Args:
            func (Callable[..., JsonType]): The function to wrap
            config (WrapperConfig | None, optional): The configuration for the wrapper.
        """
        self.func = func
        self.config = config or WrapperConfig()

    @property
    def parsers(self) -> list[Type[ArgSchemaParser]]:
        """Get the parsers for this function

        Returns:
            list[Type[ArgSchemaParser]]: The parsers
        """
        return self.config.parsers or defargparsers

    @property
    def save_return(self) -> bool:
        """Get whether to send the return value of this function to the AI

        Returns:
            bool: Whether to send the return value to the AI
        """
        return self.config.save_return

    @property
    def serialize(self) -> bool:
        """Get whether to serialize the return value of this function

        The function should return strictly a string if this is false.

        Returns:
            bool: Whether to serialize the return value
        """
        return self.config.serialize

    @property
    def interpret_as_response(self) -> bool:
        """Get whether to interpret the return value as an assistant response

        Returns:
            bool: Whether to interpret the return value as a response
        """
        return self.config.interpret_as_response

    @property
    def argument_parsers(self) -> OrderedDict[str, ArgSchemaParser]:
        """Get the argument parsers for this function

        Returns:
            OrderedDict[str, ArgSchemaParser]: The argument parsers
        """
        return OrderedDict(
            (name, self.parse_argument(argument))
            for name, argument in inspect.signature(self.func).parameters.items()
        )

    @property
    def required_arguments(self) -> JsonType:
        """Get the required arguments for this function

        Returns:
            JsonType: The required arguments
        """
        return [
            name
            for name, argument in inspect.signature(self.func).parameters.items()
            if argument.default is argument.empty
        ]

    @property
    def arguments_schema(self) -> JsonType:
        """Get the arguments schema for this function

        Returns:
            JsonType: The arguments schema
        """
        return {
            name: {
                **parser.argument_schema,
                **(
                    {"description": self.arg_docs.get(name)}
                    if name in self.arg_docs
                    else {}
                ),
            }
            for name, parser in self.argument_parsers.items()
        }

    @property
    def parsed_docs(self) -> Docstring:
        """Get the parsed docs for this function

        Returns:
            Docstring: The parsed docs
        """
        return parse(self.func.__doc__ or "")

    @property
    def arg_docs(self) -> dict[str, str]:
        """Get the argument docs for this function

        Returns:
            dict[str, str]: The argument docs
        """
        return {
            param.arg_name: param.description
            for param in self.parsed_docs.params
            if param.description
        }

    @property
    def name(self) -> str:
        """Get the name of this function

        Returns:
            str: The name
        """
        return self.func.__name__

    @property
    def schema(self) -> dict[str, JsonType]:
        """Get the schema for this function

        Returns:
            dict[str, JsonType]: The schema
        """
        return {
            "name": self.name,
            "description": self.parsed_docs.short_description,
            "parameters": {
                "type": "object",
                "properties": self.arguments_schema,
                "required": self.required_arguments,
            },
        }

    def parse_argument(self, argument: inspect.Parameter) -> ArgSchemaParser:
        """Parse an argument

        Args:
            argument (inspect.Parameter): The argument to parse

        Raises:
            TypeError: If the argument cannot be parsed

        Returns:
            ArgSchemaParser: The parser for the argument
        """
        # The reasoning behind not using pydantic is OpenAI's apparent inability to
        # parse JSON Schemas with $ref's in them - or at least, that's what I've
        # gathered from the error messages.
        for parser in self.parsers:
            if parser.can_parse(argument.annotation):
                return parser(argument.annotation, self.parsers)
        raise TypeError(f"Cannot parse argument {argument}")

    def parse_arguments(self, arguments: dict[str, JsonType]) -> OrderedDict[str, Any]:
        """Parse arguments

        Args:
            arguments (dict[str, JsonType]): The arguments to parse

        Returns:
            OrderedDict[str, Any]: The parsed arguments
        """
        return OrderedDict(
            (name, self.argument_parsers[name].parse_value(value))
            for name, value in arguments.items()
        )

    def __call__(self, arguments: dict[str, JsonType]) -> JsonType:
        """Call the wrapped function

        Args:
            arguments (dict[str, JsonType]): The arguments to call the function with

        Returns:
            JsonType: The result of the function
        """
        return self.func(**self.parse_arguments(arguments))



"""Parser for union types"""
from __future__ import annotations
import contextlib

from .abc import ArgSchemaParser

try:
    from types import UnionType
    from typing import Any, TYPE_CHECKING, Type, Union, get_args, get_origin
except ImportError:
    # This is for Python 3.8
    from typing import (  # type: ignore
        Any,
        TYPE_CHECKING,
        Type,
        Union,
        get_args,
        get_origin,
        _GenericAlias as UnionType,
    )

if TYPE_CHECKING:
    from ..json_type import JsonType
    from typing_extensions import TypeGuard


class UnionParser(ArgSchemaParser[UnionType]):
    """Parser for union types"""

    @classmethod
    def can_parse(cls, argtype: Any) -> TypeGuard[Type[UnionType]]:
        return get_origin(argtype) is Union

    @property
    def argument_schema(self) -> dict[str, JsonType]:
        return {
            "anyOf": [self.parse_rec(t).argument_schema for t in get_args(self.argtype)]
        }

    def parse_value(self, value: JsonType) -> UnionType:
        for single_type in get_args(self.argtype):
            with contextlib.suppress(TypeError):
                return self.parse_rec(single_type).parse_value(value)
        raise TypeError(f"Expected one of {get_args(self.argtype)}, got {value}")



"""Parser for float types"""
from .atomic_type_parser import AtomicParser


class FloatParser(AtomicParser[float]):
    """Parser for float types"""

    _type = float
    schema_type_name: str = "number"



"""A function set that's a union of other function sets."""
from __future__ import annotations
import contextlib
from typing import TYPE_CHECKING

from .basic_set import BasicFunctionSet
from .exceptions import FunctionNotFoundError

if TYPE_CHECKING:
    from ..json_type import JsonType
    from ..openai_types import FunctionCall
    from .functions import FunctionResult
    from .sets import FunctionSet


class UnionSkillSet(BasicFunctionSet):
    """A function set that's a union of other function sets."""

    def __init__(self, *sets: FunctionSet) -> None:
        self.sets = list(sets)
        super().__init__()

    @property
    def functions_schema(self) -> list[JsonType]:
        """Get the combined functions schema

        Returns:
            list[JsonType]: The combined functions schema
        """
        return super().functions_schema + sum(
            (function_set.functions_schema for function_set in self.sets), []
        )

    def run_function(self, input_data: FunctionCall) -> FunctionResult:
        """Run the function

        Args:
            input_data (FunctionCall): The function call

        Returns:
            FunctionResult: The function output

        Raises:
            FunctionNotFoundError: If the function is not found
        """
        for function_set in self.sets:
            with contextlib.suppress(FunctionNotFoundError):
                return function_set.run_function(input_data)
        return super().run_function(input_data)

    def add_skill(self, skill: FunctionSet) -> None:
        """Add a skill

        Args:
            skill (FunctionSet): The skill
        """
        self.sets.append(skill)



"""Default parsers for ArgSchema."""
from __future__ import annotations
from typing import TYPE_CHECKING, Type

from .bool_parser import BoolParser
from .dataclass_parser import DataclassParser
from .dict_parser import DictParser
from .enum_parser import EnumParser
from .float_parser import FloatParser
from .int_parser import IntParser
from .list_parser import ListParser
from .none_parser import NoneParser
from .str_parser import StringParser
from .union_parser import UnionParser

if TYPE_CHECKING:
    from .abc import ArgSchemaParser


defargparsers: list[Type[ArgSchemaParser]] = [
    BoolParser,
    DataclassParser,
    DictParser,
    EnumParser,
    FloatParser,
    IntParser,
    ListParser,
    NoneParser,
    StringParser,
    UnionParser,
]



"""A module for running OpenAI functions"""
from __future__ import annotations
from abc import ABC, abstractmethod
from functools import partial
from typing import Callable, TYPE_CHECKING, overload

from .functions import FunctionResult, OpenAIFunction
from .wrapper import FunctionWrapper, WrapperConfig

if TYPE_CHECKING:
    from ..json_type import JsonType
    from ..openai_types import FunctionCall


class FunctionSet(ABC):
    """A skill set"""

    @property
    @abstractmethod
    def functions_schema(self) -> list[JsonType]:
        """Get the functions schema"""

    @abstractmethod
    def run_function(self, input_data: FunctionCall) -> FunctionResult:
        """Run the function

        Args:
            input_data (FunctionCall): The function call

        Raises:
            FunctionNotFoundError: If the function is not found
        """

    @abstractmethod
    def _add_function(self, function: OpenAIFunction) -> None:
        ...

    @overload
    def add_function(self, function: OpenAIFunction) -> OpenAIFunction:
        ...

    @overload
    def add_function(
        self,
        function: Callable[..., JsonType],
        *,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> Callable[..., JsonType]:
        ...

    @overload
    def add_function(
        self,
        *,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> Callable[[Callable[..., JsonType]], Callable[..., JsonType]]:
        ...

    def add_function(
        self,
        function: OpenAIFunction | Callable[..., JsonType] | None = None,
        *,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> (
        Callable[[Callable[..., JsonType]], Callable[..., JsonType]]
        | Callable[..., JsonType]
    ):
        """Add a function

        Args:
            function (OpenAIFunction | Callable[..., JsonType]): The function
            save_return (bool): Whether to send the return value of this
                function to the AI. Defaults to True.
            serialize (bool): Whether to serialize the return value of this
                function. Defaults to True. Otherwise, the return value must be a
                string.
            interpret_as_response (bool): Whether to interpret the return
                value of this function as a response of the agent. Defaults to False.

        Returns:
            Callable[[Callable[..., JsonType]], Callable[..., JsonType]]: A decorator
            Callable[..., JsonType]: The original function
        """
        if isinstance(function, OpenAIFunction):
            self._add_function(function)
            return function
        if callable(function):
            self._add_function(
                FunctionWrapper(
                    function,
                    WrapperConfig(None, save_return, serialize, interpret_as_response),
                )
            )
            return function

        return partial(
            self.add_function,
            save_return=save_return,
            serialize=serialize,
            interpret_as_response=interpret_as_response,
        )

    @abstractmethod
    def _remove_function(self, name: str) -> None:
        ...

    def remove_function(
        self, function: str | OpenAIFunction | Callable[..., JsonType]
    ) -> None:
        """Remove a function

        Args:
            function (str | OpenAIFunction | Callable[..., JsonType]): The function
        """
        if isinstance(function, str):
            self._remove_function(function)
            return
        if isinstance(function, OpenAIFunction):
            self._remove_function(function.name)
            return
        self._remove_function(function.__name__)



"""A module for type definitions for OpenAI API responses"""
from __future__ import annotations
from typing import (
    Literal,
    Protocol,
    TYPE_CHECKING,
    TypedDict,
    Union,
    overload,
    runtime_checkable,
)

if TYPE_CHECKING:
    from typing_extensions import TypeGuard


class FunctionCall(TypedDict):
    """A type for OpenAI function calls"""

    name: str
    arguments: str


class FinalResponseMessageType(TypedDict):
    """A type for OpenAI messages that are final responses"""

    role: Literal["assistant"]
    content: str


class ContentfulMessageType(TypedDict):
    """A type for OpenAI messages that are contentful"""

    role: Literal["system", "user", "assistant"]
    content: str


class IntermediateResponseMessageType(TypedDict):
    """A type for OpenAI messages that are intermediate responses"""

    role: Literal["assistant"]
    content: None
    function_call: FunctionCall


NonFunctionMessageType = Union[ContentfulMessageType, IntermediateResponseMessageType]


class FunctionMessageType(TypedDict):
    """A type for OpenAI messages"""

    role: Literal["function"]
    name: str
    content: str | None


MessageType = Union[NonFunctionMessageType, FunctionMessageType]


class Message:
    """A container for OpenAI messages"""

    @overload
    def __init__(self, message: MessageType) -> None:
        ...

    @overload
    def __init__(self, message: str) -> None:
        ...

    @overload
    def __init__(
        self, message: str, role: Literal["system", "user", "assistant"]
    ) -> None:
        ...

    def __init__(
        self,
        message: MessageType | str,
        role: Literal["system", "user", "assistant"] = "user",
    ):
        self.message: MessageType = (
            {"role": role, "content": message} if isinstance(message, str) else message
        )

    @property
    def content(self) -> str | None:
        """Get the content of the message

        Returns:
            str | None: The content of the message
        """
        return self.message["content"]

    @property
    def role(self) -> Literal["system", "user", "assistant", "function"]:
        """Get the role of the message

        Returns:
            Literal["system", "user", "assistant", "function"]: The role of the message
        """
        return self.message["role"]

    @property
    def is_function_call(self) -> bool:
        """Check if the message is a function call

        Returns:
            bool: Whether the message is a function call
        """
        return self.role == "assistant" and "function_call" in self.message

    @property
    def function_call(self) -> FunctionCall | None:
        """Get the function call

        Returns:
            FunctionCall | None: The function call
        """
        if self.message["role"] == "assistant":
            if self.message.get("content") is not None:
                return None
            return self.message.get("function_call")
        return None

    @property
    def is_final_response(self) -> bool:
        """Check if the message is a final response

        Returns:
            bool: Whether the message is a final response
        """
        return self.role == "assistant" and self.content is not None

    def as_dict(self) -> MessageType:
        """Get the message as a dictionary

        Returns:
            MessageType: The message
        """
        return self.message

    def __repr__(self) -> str:
        if self.is_function_call:
            return f"FunctionCall({self.function_call!r})"
        return f"Message({self.message['content']!r}, {self.message['role']!r})"

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Message)
            and self.content == other.content
            and self.role == other.role
            and self.function_call == other.function_call
        )

    def __hash__(self) -> int:
        return hash((self.content, self.role))


@runtime_checkable
class GenericMessage(Protocol):
    """A container for OpenAI messages"""

    message: MessageType

    @overload
    def __init__(self, message: MessageType) -> None:
        ...

    @overload
    def __init__(self, message: str) -> None:
        ...

    @overload
    def __init__(
        self, message: str, role: Literal["system", "user", "assistant"]
    ) -> None:
        ...

    def __init__(
        self,
        message: MessageType | str,
        role: Literal["system", "user", "assistant"] = "user",
    ):
        ...

    @property
    def content(self) -> str | None:
        """Get the content of the message"""

    @property
    def role(self) -> Literal["system", "user", "assistant", "function"]:
        """Get the role of the message"""
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def is_function_call(self) -> bool:
        """Check if the message is a function call"""
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def function_call(self) -> FunctionCall | None:
        """Get the function call"""

    @property
    def is_final_response(self) -> bool:
        """Check if the message is a final response"""
        ...  # pylint: disable=unnecessary-ellipsis

    def as_dict(self) -> MessageType:
        """Get the message as a dictionary"""
        ...  # pylint: disable=unnecessary-ellipsis

    def __hash__(self) -> int:
        ...


class FinalResponseMessage(GenericMessage, Protocol):
    """A container for OpenAI final response messages"""

    message: FinalResponseMessageType  # type: ignore

    @property
    def content(self) -> str:
        """Get the content of the message"""
        # This ellipsis is for Pyright #2758
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def function_call(self) -> None:
        """Get the function call"""

    @property
    def is_final_response(self) -> Literal[True]:
        """Check if the message is a final response"""
        ...  # pylint: disable=unnecessary-ellipsis


def is_final_response_message(
    message: GenericMessage,
) -> TypeGuard[FinalResponseMessage]:
    """Check if a message is a final response message

    Args:
        message (GenericMessage): The message to check

    Returns:
        TypeGuard[FinalResponseMessage]: Whether the message is a final response message
    """
    return message.is_final_response



"""Parser for atomic json types"""
from __future__ import annotations
from abc import abstractmethod
from typing import Any, TYPE_CHECKING, Type, TypeVar

from .abc import ArgSchemaParser

if TYPE_CHECKING:
    from ..json_type import JsonType
    from typing_extensions import TypeGuard

T = TypeVar("T")


class AtomicParser(ArgSchemaParser[T]):
    """Parser for atomic json values"""

    _type: Type[T]

    @property
    @abstractmethod
    def schema_type_name(self) -> str:
        """Name of the type in the json schema"""

    @classmethod
    def can_parse(cls, argtype: Any) -> TypeGuard[Type[T]]:
        return argtype is cls._type

    @property
    def argument_schema(self) -> dict[str, JsonType]:
        return {
            "type": self.schema_type_name,
        }

    def parse_value(self, value: JsonType) -> T:
        if not isinstance(value, self._type):
            raise TypeError(f"Expected {self._type()}, got {type(value)}")
        return value



"""A module for running OpenAI functions"""
from __future__ import annotations
from dataclasses import dataclass
from typing import Protocol, TYPE_CHECKING, runtime_checkable

if TYPE_CHECKING:
    from ..json_type import JsonType


@runtime_checkable
class OpenAIFunction(Protocol):
    """A protocol for OpenAI functions"""

    def __call__(self, arguments: dict[str, JsonType]) -> JsonType:
        ...

    @property
    def schema(self) -> JsonType:
        """Get the schema for this function"""

    @property
    def name(self) -> str:
        """Get the name of this function"""
        # This ellipsis is for Pyright #2758
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def save_return(self) -> bool:
        """Get whether to save the return value of this function"""
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def serialize(self) -> bool:
        """Get whether to continue running after this function"""
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def interpret_as_response(self) -> bool:
        """Get whether to interpret the return value of this function as a response"""
        ...  # pylint: disable=unnecessary-ellipsis


@dataclass
class FunctionResult:
    """A result of a function's execution"""

    name: str
    content: str | None
    substitute: bool = False
